import yaml

from pydrodynamics.utils import EnvironmentParams, ForceCoefficients, PhysicalParams, ElectricalParams, ThrusterData, ThrusterParams, Position, Velocity

class ParamsManager:
    """
        Class to manage and load vehicle parameters from a YAML file.
    """
    def __init__(self, path):
        self.key_map = {
            'gravity': 'env.gravity',
            'density': 'env.density',

            'mass': 'physical.mass',
            'volume': 'physical.volume',

            'xg': 'physical.com.x',
            'yg': 'physical.com.y',
            'zg': 'physical.com.z',

            'xb': 'physical.cob.x',
            'yb': 'physical.cob.y',
            'zb': 'physical.cob.z',

            'ixx': 'physical.inertia.x',
            'iyy': 'physical.inertia.y',
            'izz': 'physical.inertia.z',

            'apx': 'physical.projected_area.x',
            'apy': 'physical.projected_area.y',
            'apz': 'physical.projected_area.z',

            'drag_xu': 'physical.drag.x.u',
            'drag_xv': 'physical.drag.x.v',
            'drag_xw': 'physical.drag.x.w',
            'drag_xp': 'physical.drag.x.p',
            'drag_xq': 'physical.drag.x.q',
            'drag_xr': 'physical.drag.x.r',

            'drag_yu': 'physical.drag.y.u',
            'drag_yv': 'physical.drag.y.v',
            'drag_yw': 'physical.drag.y.w',
            'drag_yp': 'physical.drag.y.p',
            'drag_yq': 'physical.drag.y.q',
            'drag_yr': 'physical.drag.y.r',

            'drag_zu': 'physical.drag.z.u',
            'drag_zv': 'physical.drag.z.v',
            'drag_zw': 'physical.drag.z.w',
            'drag_zp': 'physical.drag.z.p',
            'drag_zq': 'physical.drag.z.q',
            'drag_zr': 'physical.drag.z.r',

            'drag_ku': 'physical.drag.k.u',
            'drag_kv': 'physical.drag.k.v',
            'drag_kw': 'physical.drag.k.w',
            'drag_kp': 'physical.drag.k.p',
            'drag_kq': 'physical.drag.k.q',
            'drag_kr': 'physical.drag.k.r',

            'drag_mu': 'physical.drag.m.u',
            'drag_mv': 'physical.drag.m.v',
            'drag_mw': 'physical.drag.m.w',
            'drag_mp': 'physical.drag.m.p',
            'drag_mq': 'physical.drag.m.q',
            'drag_mr': 'physical.drag.m.r',

            'drag_nu': 'physical.drag.n.u',
            'drag_nv': 'physical.drag.n.v',
            'drag_nw': 'physical.drag.n.w',
            'drag_np': 'physical.drag.n.p',
            'drag_nq': 'physical.drag.n.q',
            'drag_nr': 'physical.drag.n.r',

            'am_xu': 'physical.added_mass.x.u',
            'am_xv': 'physical.added_mass.x.v',
            'am_xw': 'physical.added_mass.x.w',
            'am_xp': 'physical.added_mass.x.p',
            'am_xq': 'physical.added_mass.x.q',
            'am_xr': 'physical.added_mass.x.r',

            'am_yu': 'physical.added_mass.y.u',
            'am_yv': 'physical.added_mass.y.v',
            'am_yw': 'physical.added_mass.y.w',
            'am_yp': 'physical.added_mass.y.p',
            'am_yq': 'physical.added_mass.y.q',
            'am_yr': 'physical.added_mass.y.r',

            'am_zu': 'physical.added_mass.z.u',
            'am_zv': 'physical.added_mass.z.v',
            'am_zw': 'physical.added_mass.z.w',
            'am_zp': 'physical.added_mass.z.p',
            'am_zq': 'physical.added_mass.z.q',
            'am_zr': 'physical.added_mass.z.r',

            'am_ku': 'physical.added_mass.k.u',
            'am_kv': 'physical.added_mass.k.v',
            'am_kw': 'physical.added_mass.k.w',
            'am_kp': 'physical.added_mass.k.p',
            'am_kq': 'physical.added_mass.k.q',
            'am_kr': 'physical.added_mass.k.r',

            'am_mu': 'physical.added_mass.m.u',
            'am_mv': 'physical.added_mass.m.v',
            'am_mw': 'physical.added_mass.m.w',
            'am_mp': 'physical.added_mass.m.p',
            'am_mq': 'physical.added_mass.m.q',
            'am_mr': 'physical.added_mass.m.r',

            'am_nu': 'physical.added_mass.n.u',
            'am_nv': 'physical.added_mass.n.v',
            'am_nw': 'physical.added_mass.n.w',
            'am_np': 'physical.added_mass.n.p',
            'am_nq': 'physical.added_mass.n.q',
            'am_nr': 'physical.added_mass.n.r',

            'voltage': 'electrical.voltage',
            'capacity': 'electrical.capacity',
        }

        self.params_folder = "/".join(path.split('/')[:-1]) + '/'
        self.load(path)

    def load(self, path):
        with open(path, 'r') as f:
            raw = yaml.safe_load(f)
        self.name = raw['name']
        self.verbose = raw['verbose']
        self.env = EnvironmentParams(**raw['env'])
        self.physical = PhysicalParams(
            mass=raw['physical']['mass'],
            volume=raw['physical']['volume'],
            com=Position(
                x=raw['physical']['com']['x'],
                y=raw['physical']['com']['y'],
                z=raw['physical']['com']['z']
            ),
            cob=Position(
                x=raw['physical']['cob']['x'],
                y=raw['physical']['cob']['y'],
                z=raw['physical']['cob']['z']
            ),
            inertia=Position(
                x=raw['physical']['inertia']['x'],
                y=raw['physical']['inertia']['y'],
                z=raw['physical']['inertia']['z']
            ),
            projected_area=Position(
                x=raw['physical']['projected_area']['x'],
                y=raw['physical']['projected_area']['y'],
                z=raw['physical']['projected_area']['z']
            ),
            drag=ForceCoefficients(
                x=Velocity(
                    u=raw['physical']['drag']['x']['u'],
                    v=raw['physical']['drag']['x']['v'],
                    w=raw['physical']['drag']['x']['w'],
                    p=raw['physical']['drag']['x']['p'],
                    q=raw['physical']['drag']['x']['q'],
                    r=raw['physical']['drag']['x']['r']
                ),
                y=Velocity(
                    u=raw['physical']['drag']['y']['u'],
                    v=raw['physical']['drag']['y']['v'],
                    w=raw['physical']['drag']['y']['w'],
                    p=raw['physical']['drag']['y']['p'],
                    q=raw['physical']['drag']['y']['q'],
                    r=raw['physical']['drag']['y']['r']
                ),
                z=Velocity(
                    u=raw['physical']['drag']['z']['u'],
                    v=raw['physical']['drag']['z']['v'],
                    w=raw['physical']['drag']['z']['w'],
                    p=raw['physical']['drag']['z']['p'],
                    q=raw['physical']['drag']['z']['q'],
                    r=raw['physical']['drag']['z']['r']
                ),
                k=Velocity(
                    u=raw['physical']['drag']['k']['u'],
                    v=raw['physical']['drag']['k']['v'],
                    w=raw['physical']['drag']['k']['w'],
                    p=raw['physical']['drag']['k']['p'],
                    q=raw['physical']['drag']['k']['q'],
                    r=raw['physical']['drag']['k']['r']
                ),
                m=Velocity(
                    u=raw['physical']['drag']['m']['u'],
                    v=raw['physical']['drag']['m']['v'],
                    w=raw['physical']['drag']['m']['w'],
                    p=raw['physical']['drag']['m']['p'],
                    q=raw['physical']['drag']['m']['q'],
                    r=raw['physical']['drag']['m']['r']
                ),
                n=Velocity(
                    u=raw['physical']['drag']['n']['u'],
                    v=raw['physical']['drag']['n']['v'],
                    w=raw['physical']['drag']['n']['w'],
                    p=raw['physical']['drag']['n']['p'],
                    q=raw['physical']['drag']['n']['q'],
                    r=raw['physical']['drag']['n']['r']
                ),
            ),
            added_mass=ForceCoefficients(
                x=Velocity(
                    u=raw['physical']['added_mass']['x']['u'],
                    v=raw['physical']['added_mass']['x']['v'],
                    w=raw['physical']['added_mass']['x']['w'],
                    p=raw['physical']['added_mass']['x']['p'],
                    q=raw['physical']['added_mass']['x']['q'],
                    r=raw['physical']['added_mass']['x']['r']
                ),
                y=Velocity(
                    u=raw['physical']['added_mass']['y']['u'],
                    v=raw['physical']['added_mass']['y']['v'],
                    w=raw['physical']['added_mass']['y']['w'],
                    p=raw['physical']['added_mass']['y']['p'],
                    q=raw['physical']['added_mass']['y']['q'],
                    r=raw['physical']['added_mass']['y']['r']
                ),
                z=Velocity(
                    u=raw['physical']['added_mass']['z']['u'],
                    v=raw['physical']['added_mass']['z']['v'],
                    w=raw['physical']['added_mass']['z']['w'],
                    p=raw['physical']['added_mass']['z']['p'],
                    q=raw['physical']['added_mass']['z']['q'],
                    r=raw['physical']['added_mass']['z']['r']
                ),
                k=Velocity(
                    u=raw['physical']['added_mass']['k']['u'],
                    v=raw['physical']['added_mass']['k']['v'],
                    w=raw['physical']['added_mass']['k']['w'],
                    p=raw['physical']['added_mass']['k']['p'],
                    q=raw['physical']['added_mass']['k']['q'],
                    r=raw['physical']['added_mass']['k']['r']
                ),
                m=Velocity(
                    u=raw['physical']['added_mass']['m']['u'],
                    v=raw['physical']['added_mass']['m']['v'],
                    w=raw['physical']['added_mass']['m']['w'],
                    p=raw['physical']['added_mass']['m']['p'],
                    q=raw['physical']['added_mass']['m']['q'],
                    r=raw['physical']['added_mass']['m']['r']
                ),
                n=Velocity(
                    u=raw['physical']['added_mass']['n']['u'],
                    v=raw['physical']['added_mass']['n']['v'],
                    w=raw['physical']['added_mass']['n']['w'],
                    p=raw['physical']['added_mass']['n']['p'],
                    q=raw['physical']['added_mass']['n']['q'],
                    r=raw['physical']['added_mass']['n']['r']
                ),
            ),
        )
        self.electrical = ElectricalParams(**raw['electrical'])
        self.thrusters = ThrusterParams(
            data=raw['thrusters']['data'],
            list=[ThrusterData(
                name=t['name'],
                pos=Position(t['pos']['x'], t['pos']['y'], t['pos']['z']),
                dir=Position(t['dir']['x'], t['dir']['y'], t['dir']['z'])
            ) for t in raw['thrusters']['list']]
        )

        # Add thrusters to key_map by name and by index
        for i, t in enumerate(raw['thrusters']['list']):
            name = t['name']
            self.key_map[f'{name}_x'] = f'thrusters.list.{i}.pos.x'
            self.key_map[f'{name}_y'] = f'thrusters.list.{i}.pos.y'
            self.key_map[f'{name}_z'] = f'thrusters.list.{i}.pos.z'
            self.key_map[f'{name}_dir_x'] = f'thrusters.list.{i}.dir.x'
            self.key_map[f'{name}_dir_y'] = f'thrusters.list.{i}.dir.y'
            self.key_map[f'{name}_dir_z'] = f'thrusters.list.{i}.dir.z'

            self.key_map[f't{i + 1}_x'] = f'thrusters.list.{i}.pos.x'
            self.key_map[f't{i + 1}_y'] = f'thrusters.list.{i}.pos.y'
            self.key_map[f't{i + 1}_z'] = f'thrusters.list.{i}.pos.z'
            self.key_map[f't{i + 1}_dir_x'] = f'thrusters.list.{i}.dir.x'
            self.key_map[f't{i + 1}_dir_y'] = f'thrusters.list.{i}.dir.y'
            self.key_map[f't{i + 1}_dir_z'] = f'thrusters.list.{i}.dir.z'

        if self.verbose: print(f'Successfully loaded params for vehicle {self.name} from {path}')

    def get(self, key):
        """Utility function to get a deep nested attribute from params."""
        if key not in self.key_map:
            raise KeyError(f'Key {key} not found in params')
        keys = self.key_map[key].split('.')
        value = self
        for k in keys:
            if str.isnumeric(k):
                k = int(k)
                value = value[k]
            else:
                value = getattr(value, k)
        return value


import numpy as np

from pydrodynamics.params import ParamsManager
from pydrodynamics.utils import body_to_world, fix_angular_circularity, state_array_to_object, state_object_to_array

class Dynamics:
    """
        Given the external forces and moments acting on the vehicle, this class calculates the next state of the vehicle.
        Uses Runge-Kutta method to integrate the equations of motion, obtaining position and velocity.
    """
    def __init__(self, params: ParamsManager):
        self.params = params
        self.num_states = 13

        # Construct mass matrix
        # TODO: add added mass coefficients
        m = params.get('mass')
        xg = params.get('xg')
        yg = params.get('yg')
        zg = params.get('zg')
        ixx = params.get('ixx')
        iyy = params.get('iyy')
        izz = params.get('izz')
        am_xu = params.get('am_xu')
        am_yv = params.get('am_yv')
        am_yr = params.get('am_yr')
        am_zw = params.get('am_zw')
        am_zq = params.get('am_zq')
        am_kp = params.get('am_kp')
        am_mw = params.get('am_mw')
        am_mq = params.get('am_mq')
        am_nv = params.get('am_nv')
        am_nr = params.get('am_nr')
        self.inv_mass_matrix = np.linalg.inv(np.array([
            [m-am_xu,	0.0,		0.0, 			0.0, 		m*zg, 		    -m*yg],
            [0.0,       m-am_yv,    0.0,       		-m*zg,     	0.0,       	    m*xg-am_yr],
            [0.0,       0.0,       	m-am_zw,    	m*yg,  		-m*xg-am_zq,    0.0],
            [0.0,      	-m*zg,      m*yg,    		ixx-am_kp,  0.0,            0.0],
            [m*zg,      0.0,    	-m*xg-am_mw, 	0.0,     	iyy-am_mq,      0.0],
            [-m*yg,  	m*xg-am_nv, 0.0,        	0.0,       	0.0,     	    izz-am_nr]
        ]))

    def calculate(self, dt, state, tau):
        """Calculate the next state of the vehicle based on external forces."""
        if self.params.verbose: print(tau)

        # Create a new state in np array format
        state = state_object_to_array(state)

        # Calculate the next state using Runge-Kutta method
        next_state = self.runge_kutta(dt, state, tau)
        if self.params.verbose: print(next_state)

        # Convert the next state back to State object
        return state_array_to_object(next_state)

    def runge_kutta(self, dt, state, tau):
        """
            Runge-Kutta method to solve the dynamics of the vehicle accurately.
            Integrates and approximates the result from solve_dynamics over time.
            
            Input:
                dt: Time step for the simulation.
                state: Current state of the vehicle, in np array format.
                tau: External force and moment due to thrusters in body fixed frame.
            Output:
                next_state: The next state of the vehicle after applying the thruster forces, in np array format. 
        """
        k1 = self.solve_dynamics(state, tau)
        k2 = self.solve_dynamics(state + k1 * dt / 2, tau)
        k3 = self.solve_dynamics(state + k2 * dt / 2, tau)
        k4 = self.solve_dynamics(state + k3 * dt, tau)

        next_state = state + (k1 + 2*k2 + 2*k3 + k4) * dt / 6
        return fix_angular_circularity(next_state)

    def solve_dynamics(self, state, tau):
        """
            Solves the kinematics and dynamics of the vehicle, but needs runge-kutta since these are
            non-linear and coupled ODEs. Runge-Kutta will approximate the state over time for a more precise result.

            Input:
                dt: Time step for the simulation.
                state: Current state of the vehicle, in np array format.
                tau: External force and moment due to thrusters in body fixed frame.
            Output:
                delta_state: The change in state of the vehicle after applying the thruster forces, in np array format, denoting acceleration and new velocities.
        """
        # Unpack state
        u, v, w, p, q, r = state[6:12]

        # Solve for accelerations
        accels = np.matmul(self.inv_mass_matrix, tau).reshape((6,))

        # Convert body-fixed velocities to world frame to match world fixed frame for position and orientation when integrating
        r_trans, r_rot = body_to_world(state)
        vels = np.concatenate((
            np.matmul(r_trans, np.array([u, v, w])),
            np.matmul(r_rot, np.array([p, q, r]))
        ), axis=0).reshape((6,))

        return np.concatenate((vels, accels, [0]), axis=0).reshape((self.num_states,))



from pydrodynamics.utils import State, Position, Orientation, LinearVelocity, AngularVelocity
from pydrodynamics.dynamics import Dynamics
from pydrodynamics.params import ParamsManager
from pydrodynamics.thrusters import Thrusters
from pydrodynamics.environment import Environment

class Vehicle:
	"""
		Main class tying all the different modules together.
		It initializes the vehicle from the given yaml file and calculates the next state of the vehicle given PWM control signals.
	"""
	def __init__(self, params, initial_state: State = None):
		self.params = ParamsManager(params)
		self.dynamics = Dynamics(self.params)
		self.thrusters = Thrusters(self.params)
		self.env = Environment(self.params)

		self.state = initial_state if initial_state else State(
			position=Position(0, 0, 0),
			orientation=Orientation(0, 0, 0),
			linear_velocity=LinearVelocity(0, 0, 0),
			angular_velocity=AngularVelocity(0, 0, 0),
			voltage=self.params.get('voltage'),
		)

	def step(self, dt, pwm_array, next_state=None):
		"""
			Calculate the next state of the vehicle.

			Input:
				dt: Time step for the simulation.
				pwm_array: Array of PWM values for the thrusters.
				next_state: If given, this will override the current state, used for collision detection.
			Output:
				next_state: The next state of the vehicle after applying the thruster forces.
		"""
		# Calculate total external forces and moments
		tau_thrusters = self.thrusters.calculate(self.state, pwm_array)
		tau_env = self.env.calculate(self.state)
		tau = tau_thrusters + tau_env

		# Calculate the next state using the dynamics model
		self.state = next_state if next_state else self.dynamics.calculate(dt, self.state, tau)
		print(self.state)
		return self.state



import numpy as np

from pydrodynamics.params import ParamsManager
from pydrodynamics.utils import unpack_state_object
from scipy.interpolate import RegularGridInterpolator

class Thrusters:
    """
        Find thruster forces and moments based on PWM values.
        Uses thruster data from the given CSV file and interpolates with the current voltage.
        Also takes into account the thruster positions and directions used in the vehicle model.
    """
    def __init__(self, params: ParamsManager):
        self.params = params

        self.num_thrusters = len(params.thrusters.list)
        self.tam = np.zeros((6, self.num_thrusters))
        self.thruster_positions = np.zeros((self.num_thrusters, 3))
        self.thruster_directions = np.zeros((self.num_thrusters, 3))
        for i in range(self.num_thrusters):
            # Construct thruster allocation matrix
            self.tam[0:3, i] = np.array([params.get(f't{i + 1}_dir_x'),
                                        params.get(f't{i + 1}_dir_y'),
                                        params.get(f't{i + 1}_dir_z')])
            self.tam[3:6, i] = np.cross(np.array([params.get(f't{i + 1}_x') + params.get('xg'),
                                                params.get(f't{i + 1}_y') + params.get('yg'),
                                                params.get(f't{i + 1}_z') + params.get('zg')]),
                                        self.tam[0:3, i])
            
            # Construct thruster position and direction vectors
            self.thruster_positions[i, :] = np.array([params.get(f't{i + 1}_x') + params.get('xg'),
                                                    params.get(f't{i + 1}_y') + params.get('yg'),
                                                    params.get(f't{i + 1}_z') + params.get('zg')])
            direction_vector = np.array([params.get(f't{i + 1}_dir_x'),
                                        params.get(f't{i + 1}_dir_y'),
                                        params.get(f't{i + 1}_dir_z')])
            self.thruster_directions[i, :] = direction_vector / np.linalg.norm(direction_vector)

        # Load thruster data
        self.load_thruster_data()

    def calculate(self, state, pwm_array) -> np.ndarray:
        """
            Calculate the thrust forces based on PWM values.
            
            Input:
                pwm_array: Array of PWM values for the thrusters.
            Output:
                tau_t: External force and moment due to thrusters in body fixed frame, in (6,) format.
        """
        # Convert PWM values to thrust forces
        thrust_forces = self.pwm_to_force(state, pwm_array) * self.thruster_directions

        # Calculate thrust moments
        thrust_moments = np.cross(self.thruster_positions, thrust_forces)

        # Combine thrust forces and moments into a single vector
        tau_t = np.zeros((6,))
        tau_t[0:3] = np.sum(thrust_forces, axis=0).reshape((3,))
        tau_t[3:6] = np.sum(thrust_moments, axis=0).reshape((3,))

        return tau_t

    def load_thruster_data(self):
        """Load thruster data from the specified file under thruster data."""
        data = np.loadtxt(f"{self.params.params_folder}{self.params.thrusters.data}", delimiter=',', skiprows=1)

        # Extract the relevant columns
        self.voltage_values = np.unique(data[:, 0])  # Voltage is column A (0-indexed)
        self.pwm_values = np.unique(data[:, 1])      # PWM is column B (0-indexed)
        force = data[:, 6]                      # Force is column G (0-indexed)

        # Initialize force_values matrix
        num_rows = len(self.voltage_values)
        num_cols = len(self.pwm_values)
        self.force_values = np.zeros((num_rows, num_cols))

        # Fill in the matrix row by row
        for i in range(num_rows):
            start_idx = i * num_cols
            end_idx = (i + 1) * num_cols
            self.force_values[i, :] = force[start_idx:end_idx]

        # Convert from kgf to Newtons (1 kgf = 9.81 N)
        self.force_values *= 9.81

    def pwm_to_force(self, state, pwm_array) -> np.ndarray:
        """
            Convert PWM values to thrust forces using the thruster data.

            Input:
                pwm_array: Array of PWM values for the thrusters.
            Output:
                thrust_forces: The calculated thrust forces from each thruster.
        """
        # Unpack state
        voltage = unpack_state_object(state)[12]

        # Make sure pwm is in the range [1100, 1900]
        pwm_array = np.clip(pwm_array, 1100, 1900)

        # Create the interpolator
        interpolator = RegularGridInterpolator(
            (self.voltage_values, self.pwm_values),
            self.force_values,
            method='linear',
            bounds_error=False,
            fill_value=None
        )

        # Interpolate forces for each PWM value at the given voltage
        points = np.array([(voltage, pwm) for pwm in pwm_array])
        return interpolator(points)[:, np.newaxis]
